-- ReforgeLite v1.0 by d07.RiV (Iroared)
-- All rights reserved

local REFORGE_COEFF = 0.4
local REFORGE_CHEAT = 5
local L = ReforgeLiteLocale

local function DeepCopy (t, cache)
  if type (t) ~= "table" then
    return t
  end
  local copy = {}
  for i, v in pairs (t) do
    if type (v) ~= "table" then
      copy[i] = v
    else
      cache = cache or {}
      cache[t] = copy
      if cache[v] then
        copy[i] = cache[v]
      else
        copy[i] = DeepCopy (v, lookup)
      end
    end
  end
  return copy
end

----------------------------------------- CAP PRESETS ---------------------------------

function ReforgeLite:RatingPerPoint (stat, level)
  level = level or UnitLevel ("player")
  local factor
  if level <= 34 and (stat == self.STATS.DODGE or stat == self.STATS.PARRY) then
    factor = 0.5
  elseif level <= 10 then
    factor = 1 / 26
  elseif level <= 60 then
    factor = (level - 8) / 52
  elseif level <= 70 then
    factor = 82 / (262 - 3 * level)
  elseif level <= 80 then
    factor = (82 / 52) * math.pow (131 / 63, (level - 70) / 10)
  else
    factor = (82 / 52) * (131 / 63)
    if level == 81 then
      factor = factor * 1.31309
    elseif level == 82 then
      factor = factor * 1.72430
    elseif level == 83 then
      factor = factor * 2.26519
    elseif level == 84 then
      factor = factor * 2.97430
    elseif level == 85 then
      factor = factor * 3.90537
    end
  end
  if stat == self.STATS.DODGE or stat == self.STATS.PARRY then
    return factor * 13.8
  elseif stat == self.STATS.HIT then
    return factor * 9.37931
  elseif stat == self.STATS.SPELLHIT then
    return factor * 8
  elseif stat == self.STATS.HASTE then
    return factor * 10
  elseif stat == self.STATS.CRIT then
    return factor * 14
  elseif stat == self.STATS.EXP then
    return factor * 2.34483
  elseif stat == self.STATS.MASTERY then
    return factor * 14
  end
end
function ReforgeLite:GetMeleeHitBonus ()
  return GetHitModifier ()
end
function ReforgeLite:GetSpellHitBonus ()
  return GetSpellHitModifier ()
end
function ReforgeLite:GetExpertiseBonus ()
  return GetExpertise () - math.floor (GetCombatRatingBonus (CR_EXPERTISE))
end
function ReforgeLite:GetNeededMeleeHit ()
  local diff = self.db.targetLevel
  if diff <= 2 then
    return math.max (0, 5 + 0.5 * diff)
  else
    return 2 + 2 * diff
  end
end
function ReforgeLite:GetNeededSpellHit ()
  local diff = self.db.targetLevel
  if diff <= 2 then
    return math.max (0, 4 + diff)
  else
    return 11 * diff - 16
  end
end
function ReforgeLite:GetNeededExpertiseSoft ()
  local diff = self.db.targetLevel
  return math.ceil (math.max (0, 5 + 0.5 * diff) / 0.25)
end
function ReforgeLite:GetNeededExpertiseHard ()
  local diff = self.db.targetLevel
  if diff <= 2 then
    return math.ceil (math.max (0, 5 + 0.5 * diff) / 0.25)
  else
    return math.ceil (14 / 0.25)
  end
end

ReforgeLite.capPresets = {
  {
    value = 1,
    name = L["Manual"],
    getter = nil
  },
  {
    value = 2,
    name = L["Melee hit cap"],
    getter = function ()
      return ReforgeLite:RatingPerPoint (ReforgeLite.STATS.HIT) * (ReforgeLite:GetNeededMeleeHit () - ReforgeLite:GetMeleeHitBonus ())
    end
  },
  {
    value = 3,
    name = L["Spell hit cap"],
    getter = function ()
      return ReforgeLite:RatingPerPoint (ReforgeLite.STATS.SPELLHIT) * (ReforgeLite:GetNeededSpellHit () - ReforgeLite:GetSpellHitBonus ())
    end
  },
  {
    value = 4,
    name = L["Melee DW hit cap"],
    getter = function ()
      return ReforgeLite:RatingPerPoint (ReforgeLite.STATS.HIT) * (ReforgeLite:GetNeededMeleeHit () + 19 - ReforgeLite:GetMeleeHitBonus ())
    end
  },
  {
    value = 5,
    name = L["Expertise soft cap"],
    getter = function ()
      return ReforgeLite:RatingPerPoint (ReforgeLite.STATS.EXP) * (ReforgeLite:GetNeededExpertiseSoft () - ReforgeLite:GetExpertiseBonus ())
    end
  },
  {
    value = 6,
    name = L["Expertise hard cap"],
    getter = function ()
      return ReforgeLite:RatingPerPoint (ReforgeLite.STATS.EXP) * (ReforgeLite:GetNeededExpertiseHard () - ReforgeLite:GetExpertiseBonus ())
    end
  },
}

---------------------------------------------------------------------------------------

function ReforgeLite:GetItemReforgeOptions (item, data)
  local opt = {}
  local best = nil
  for i = 1, #self.itemStats do
    if item.stats[i] == 0 and i ~= data.caps[1].stat and i ~= data.caps[2].stat and i ~= data.spi
       and (best == nil or data.weights[i] > data.weights[best]) then
      best = i
    end
  end
  if best then
    local worst = nil
    local worstScore = 0
    for i = 1, #self.itemStats do
      if item.stats[i] > 0 and i ~= data.caps[1].stat and i ~= data.caps[2].stat and i ~= data.spi then
        local score = (data.weights[best] - data.weights[i]) * math.floor (item.stats[i] * REFORGE_COEFF)
        if score > worstScore then
          worstScore = score
          worst = i
        end
      end
    end
    if worst then
      table.insert (opt, {src = worst, dst = best, d1 = 0, d2 = 0, score = worstScore})
    else
      table.insert (opt, {d1 = 0, d2 = 0, score = 0})
    end
  else
    table.insert (opt, {d1 = 0, d2 = 0, score = 0})
  end
  for s = 1, 2 do
    if data.caps[s].stat > 0 then
      if item.stats[data.caps[s].stat] == 0 then
        for i = 1, #self.itemStats do
          if item.stats[i] > 0 and (i ~= data.spi or data.caps[s].stat ~= self.STATS.HIT) then
            local amount = math.floor (item.stats[i] * REFORGE_COEFF)
            local dststat = data.caps[s].stat
            if self.db.convertSpirit and dststat == self.STATS.HIT and data.weights[self.STATS.SPIRIT] > 0 then
              dststat = self.STATS.SPIRIT
            end
            local weight = (i == data.caps[3 - s].stat and 0 or (amount * ((dststat == data.caps[s].stat and 0 or data.weights[dststat]) - data.weights[i])))
            table.insert (opt, {src = i, dst = dststat,
              ["d" .. s] = amount, ["d" .. (3 - s)] = ((i == data.caps[3 - s].stat or i == data.spi) and -amount or 0), score = weight})
          end
        end
      elseif best then
        local amount = math.floor (item.stats[data.caps[s].stat] * REFORGE_COEFF)
        table.insert (opt, {src = data.caps[s].stat, dst = best,
          d1 = (s == 1 and -amount or 0), d2 = (s == 2 and -amount or 0), score = data.weights[best] * amount})
      end
      if self.db.convertSpirit and data.caps[s].stat == self.STATS.HIT and item.stats[data.spi] > 0 then
        if best then
          local amount = math.floor (item.stats[self.STATS.SPIRIT] * REFORGE_COEFF)
          table.insert (opt, {src = self.STATS.SPIRIT, dst = best,
            d1 = (s == 1 and -amount or 0), d2 = (s == 2 and -amount or 0), score = (data.weights[best] - data.weights[self.STATS.SPIRIT]) * amount})
        end
      end
    end
  end
  return opt
end
function ReforgeLite:FinalizeReforge (code, data)
  local method = data.method
  method.stats = DeepCopy (data.initial)
  method.weights = DeepCopy (data.weights)
  method.caps = DeepCopy (data.caps)
  for i = 1, #method.items do
    local opt = self:GetItemReforgeOptions (method.items[i], data)[code:byte (i)]
    method.items[i].src = opt.src
    method.items[i].dst = opt.dst
    method.items[i].reforge = nil
    if method.items[i].src and method.items[i].dst then
      method.items[i].amount = math.floor (method.items[i].stats[method.items[i].src] * REFORGE_COEFF)
      method.items[i].stats[method.items[i].src] = method.items[i].stats[method.items[i].src] - method.items[i].amount
      method.items[i].stats[method.items[i].dst] = method.items[i].stats[method.items[i].dst] + method.items[i].amount
      for j = 1, #self.reforgeTable do
        if self.reforgeTable[j][1] == method.items[i].src and self.reforgeTable[j][2] == method.items[i].dst then
          method.items[i].reforge = j
          break
        end
      end
    end
    for j = 1, #self.itemStats do
      method.stats[j] = method.stats[j] + method.items[i].stats[j]
    end
  end
  if self.db.convertSpirit then
    method.stats[self.STATS.HIT] = method.stats[self.STATS.HIT] + method.stats[self.STATS.SPIRIT]
  end
  method.score = 0
  for i = 1, #self.itemStats do
    local value = method.stats[i]
    method.score = method.score + self:GetStatScore (data.weights, data.caps, i, value)
  end
end
function ReforgeLite:InitReforge ()
  local method = {}
  method.items = {}
  for i = 1, #self.itemData do
    method.items[i] = {}
    method.items[i].stats = {}
    local item = GetInventoryItemLink ("player", self.itemData[i].slotId)
    local stats = (item and GetItemStats (item) or {})
    for j = 1, #self.itemStats do
      method.items[i].stats[j] = (stats[self.itemStats[j].name] or 0)
    end
  end

  REFORGE_CHEAT = self.db.reforgeCheat

  local data = {}
  data.method = method
  data.weights = DeepCopy (self.db.statWeights)
  data.caps = DeepCopy (self.db.statCaps)
  data.caps[1].init = 0
  data.caps[2].init = 0
  data.initial = {}
  for i = 1, #self.itemStats do
    data.initial[i] = self.itemStats[i].getter ()
    for j = 1, #data.method.items do
      data.initial[i] = data.initial[i] - data.method.items[j].stats[i]
    end
  end
  for i = 1, #data.method.items do
    local item = GetInventoryItemLink ("player", self.itemData[i].slotId)
    local reforge = (item and self:GetReforgeID (item))
    if reforge then
      src, dst = self.reforgeTable[reforge][1], self.reforgeTable[reforge][2]
      local amount = math.floor (method.items[i].stats[src] * REFORGE_COEFF)
      data.initial[src] = data.initial[src] + amount
      data.initial[dst] = data.initial[dst] - amount
    end
  end
  if self.db.convertSpirit then
    data.initial[self.STATS.HIT] = data.initial[self.STATS.HIT] - self.itemStats[self.STATS.SPIRIT].getter ()
  end
  if data.caps[1].stat > 0 then
    data.caps[1].init = data.initial[data.caps[1].stat]
    for i = 1, #data.method.items do
      data.caps[1].init = data.caps[1].init + data.method.items[i].stats[data.caps[1].stat]
    end
  end
  if data.caps[2].stat > 0 then
    data.caps[2].init = data.initial[data.caps[2].stat]
    for i = 1, #data.method.items do
      data.caps[2].init = data.caps[2].init + data.method.items[i].stats[data.caps[2].stat]
    end
  end
  if data.caps[1].stat == 0 then
    data.caps[1], data.caps[2] = data.caps[2], data.caps[1]
  end
  if data.caps[2].stat == data.caps[1].stat then
    data.caps[2].stat = 0
    data.caps[2].init = 0
  end

  data.spi = 0
  if self.db.convertSpirit then
    if data.caps[1].stat == self.STATS.HIT then
      data.spi = self.STATS.SPIRIT
      data.caps[1].init = data.caps[1].init + data.initial[self.STATS.SPIRIT]
      for i = 1, #data.method.items do
        data.caps[1].init = data.caps[1].init + data.method.items[i].stats[self.STATS.SPIRIT]
      end
    end
    if data.caps[2].stat == self.STATS.HIT then
      data.spi = self.STATS.SPIRIT
      data.caps[2].init = data.caps[2].init + data.initial[self.STATS.SPIRIT]
      for i = 1, #data.method.items do
        data.caps[2].init = data.caps[2].init + data.method.items[i].stats[self.STATS.SPIRIT]
      end
    end
  end

  return data
end

function ReforgeLite:CapAllows (cap, value)
  for i = 1, #cap.points do
    if cap.points[i].method == 1 and value < cap.points[i].value then
      return false
    elseif cap.points[i].method == 2 and value > cap.points[i].value then
      return false
    end
  end
  return true
end
function ReforgeLite:Compute ()
  local data = self:InitReforge ()

  local TABLE_SIZE = 10000
  local scores, codes = {}, {}
  local linit = math.floor (data.caps[1].init / REFORGE_CHEAT + math.random ()) + math.floor (data.caps[2].init / REFORGE_CHEAT + math.random ()) * TABLE_SIZE
  local reforgeOptions = {}
  scores[linit] = 0
  codes[linit] = ""
  for i = 1, #self.itemData do
    local newscores, newcodes = {}, {}
    local opt = self:GetItemReforgeOptions (data.method.items[i], data)
    reforgeOptions[i] = opt
    for k, score in pairs (scores) do
      local code = codes[k]
      local s1 = k % TABLE_SIZE
      local s2 = math.floor (k / TABLE_SIZE)
      for j = 1, #opt do
        local o = opt[j]
        local nscore = score + o.score
        local nk = s1 + math.floor (o.d1 / REFORGE_CHEAT + math.random ()) + (s2 + math.floor (o.d2 / REFORGE_CHEAT + math.random ())) * TABLE_SIZE
        if newscores[nk] == nil or nscore > newscores[nk] then
          newscores[nk] = nscore
          newcodes[nk] = code .. string.char (j)
        end
      end
    end
    scores, codes = newscores, newcodes
  end
  collectgarbage ("collect")

  local bestCode = nil
  local bestScore = 0
  for k, score in pairs (scores) do
    local s1 = data.caps[1].init
    local s2 = data.caps[2].init
    local code = codes[k]
    for i = 1, #code do
      local b = code:byte (i)
      s1 = s1 + reforgeOptions[i][b].d1
      s2 = s2 + reforgeOptions[i][b].d2
    end
    local allow = true
    if data.caps[1].stat > 0 then
      allow = allow and self:CapAllows (data.caps[1], s1)
      score = score + self:GetCapScore (data.weights, data.caps[1], s1)
    end
    if data.caps[2].stat > 0 then
      allow = allow and self:CapAllows (data.caps[2], s2)
      score = score + self:GetCapScore (data.weights, data.caps[2], s2)
    end
    if allow and (bestCode == nil or score > bestScore) then
      bestCode = code
      bestScore = score
    end
  end
  self:FinalizeReforge (bestCode, data)
  return data.method
end
